@@ -1,6 +1,6 @@
 ################################################################################
 #
-#  Copyright (C) 2002-2005  Travis Shirk <travis@pobox.com>
+#  Copyright (C) 2002-2006  Travis Shirk <travis@pobox.com>
 #  Copyright (C) 2001  Ryan Finne <ryan@finnie.org>
 #
 #  This program is free software; you can redistribute it and/or modify
@@ -44,6 +44,7 @@ SUBTITLE_FID       = "TIT3";
 CONTENT_TITLE_FID  = "TIT1";
 TITLE_FIDS         = [TITLE_FID, SUBTITLE_FID, CONTENT_TITLE_FID];
 COMMENT_FID        = "COMM";
+LYRICS_FID         = "USLT";
 GENRE_FID          = "TCON";
 TRACKNUM_FID       = "TRCK";
 DISCNUM_FID        = "TPOS";
@@ -261,6 +262,7 @@ USERTEXT_FRAME_RX = re.compile("^" + USERTEXT_FID + "$");
 URL_FRAME_RX = re.compile("^W[A-Z0-9][A-Z0-9][A-Z0-9]$");
 USERURL_FRAME_RX = re.compile("^" + USERURL_FID + "$");
 COMMENT_FRAME_RX = re.compile("^" + COMMENT_FID + "$");
+LYRICS_FRAME_RX = re.compile("^" + LYRICS_FID + "$");
 CDID_FRAME_RX = re.compile("^" + CDID_FID + "$");
 IMAGE_FRAME_RX = re.compile("^" + IMAGE_FID + "$");
 PLAYCOUNT_FRAME_RX = re.compile("^" + PLAYCOUNT_FID + "$");
@@ -973,7 +975,13 @@ class UserURLFrame(URLFrame):
       self.encoding = data[0];
       TRACE_MSG("UserURLFrame encoding: %s" %\
                 id3EncodingToString(self.encoding));
-      (d, u) = splitUnicode(data[1:], self.encoding);
+      try:
+          (d, u) = splitUnicode(data[1:], self.encoding);
+      except ValueError, ex:
+          if strictID3():
+              raise FrameException("Invalid WXXX frame, no null byte")
+          d = data[1:]
+          u = ""
       self.description = unicode(d, id3EncodingToString(self.encoding));
       TRACE_MSG("UserURLFrame description: %s" % self.description);
       self.url = u;
@@ -1073,6 +1081,87 @@ class CommentFrame(Frame):
              (self.getFrameDesc(), self.header.id, self.comment,
               self.lang, self.description);
 
+################################################################################
+class LyricsFrame(Frame):
+   lang = "";
+   description = u"";
+   lyrics = u"";
+
+   # Data string format:
+   # encoding (one byte) + lang (three byte code) + description + "\x00" +
+   # text
+   def __init__(self, frameHeader, data = None, lang = "",
+                description = u"", lyrics = u"", encoding = DEFAULT_ENCODING):
+       Frame.__init__(self, frameHeader);
+       if data != None:
+           self._set(data, frameHeader);
+       else:
+           assert(isinstance(description, unicode));
+           assert(isinstance(lyrics, unicode));
+           assert(isinstance(lang, str));
+           self.encoding = encoding;
+           self.lang = lang;
+           self.description = description;
+           self.lyrics = lyrics;
+
+   # Data string format:
+   # encoding (one byte) + lang (three byte code) + description + "\x00" +
+   # text
+   def _set(self, data, frameHeader = None):
+      assert(frameHeader);
+      if not LYRICS_FRAME_RX.match(frameHeader.id):
+         raise FrameException("Invalid frame id for LyricsFrame: " +\
+                              frameHeader.id);
+
+      data = self.disassembleFrame(data);
+      self.encoding = data[0];
+      TRACE_MSG("LyricsFrame encoding: " + id3EncodingToString(self.encoding));
+      try:
+          self.lang = str(data[1:4]).strip("\x00");
+          # Test ascii encoding
+          temp_lang = unicode(self.lang, "ascii");
+          if self.lang and \
+             not re.compile("[A-Z][A-Z][A-Z]", re.IGNORECASE).match(self.lang):
+             if strictID3():
+                 raise FrameException("[LyricsFrame] Invalid language "\
+                                       "code: %s" % self.lang);
+      except UnicodeDecodeError, ex:
+          if strictID3():
+              raise FrameException("[LyricsFrame] Invalid language code: "\
+                                   "[%s] %s" % (ex.object, ex.reason));
+          else:
+              self.lang = "";
+      try:
+         (d, c) = splitUnicode(data[4:], self.encoding);
+         self.description = unicode(d, id3EncodingToString(self.encoding));
+         self.lyrics = unicode(c, id3EncodingToString(self.encoding));
+      except ValueError:
+          if strictID3():
+              raise FrameException("Invalid lyrics; no description/lyrics");
+          else:
+              self.description = u"";
+              self.lyrics = u"";
+      if not strictID3():
+          self.description = cleanNulls(self.description)
+          self.lyrics = cleanNulls(self.lyrics)
+
+   def render(self):
+      lang = self.lang.encode("ascii");
+      if len(lang) > 3:
+          lang = lang[0:3];
+      elif len(lang) < 3:
+          lang = lang + ('\x00' * (3 - len(lang)));
+      data = self.encoding + lang +\
+             self.description.encode(id3EncodingToString(self.encoding)) +\
+             self.getTextDelim() +\
+             self.lyrics.encode(id3EncodingToString(self.encoding));
+      return self.assembleFrame(data);
+
+   def __unicode__(self):
+      return u"<%s (%s): %s [Lang: %s] [Desc: %s]>" %\
+             (self.getFrameDesc(), self.header.id, self.lyrics,
+              self.lang, self.description);
+
 ################################################################################
 # This class refers to the APIC frame, otherwise known as an "attached
 # picture".
@@ -1410,13 +1499,12 @@ class UnknownFrame(Frame):
    data = "";
 
    def __init__(self, frameHeader, data):
-       Frame.__init__(self, frameHeader);
        assert(frameHeader and data);
+       Frame.__init__(self, frameHeader);
        self._set(data, frameHeader);
 
    def _set(self, data, frameHeader):
-      data = self.disassembleFrame(data);
-      self.data = data;
+      self.data = self.disassembleFrame(data);
 
    def render(self):
       return self.assembleFrame(self.data)
@@ -1558,6 +1646,17 @@ class FrameSet(list):
                                      language and description not allowed." %\
                                      fid);
 
+      # Lyrics frame restrictions.
+      # Multiples must have a unique description/language combination.
+      if strictID3() and LYRICS_FRAME_RX.match(fid) and self[fid]:
+         lyricsFrames = self[fid];
+         for frm in lyricsFrames:
+            if frm.description == frame.description and\
+               frm.lang == frame.lang:
+               raise FrameException("Multiple %s frames with the same\
+                                     language and description not allowed." %\
+                                     fid);
+
       # URL frame restrictions.
       # No multiples except for TXXX which must have unique descriptions.
       if strictID3() and URL_FRAME_RX.match(fid) and self[fid]:
@@ -1662,6 +1761,38 @@ class FrameSet(list):
                                    description = description,
                                    comment = comment));
 
+   # If a user text frame with the same description exists then
+   # the frame text is replaced, otherwise the frame is added.
+   def setLyricsFrame(self, lyrics, description, lang = DEFAULT_LANG,
+                       encoding = None):
+      assert(isinstance(lyrics, unicode) and isinstance(description, unicode));
+
+      if self[LYRICS_FID]:
+         found = 0;
+         for f in self[LYRICS_FID]:
+            if f.lang == lang and f.description == description:
+               f.lyrics = lyrics;
+               if encoding:
+                   f.encoding = encoding;
+               found = 1;
+               break;
+         if not found:
+            h = FrameHeader(self.tagHeader);
+            h.id = LYRICS_FID;
+            if not encoding:
+                encoding = DEFAULT_ENCODING;
+            self.addFrame(LyricsFrame(h, encoding = encoding, lang = lang,
+                                       description = description,
+                                       lyrics = lyrics));
+      else:
+        if not encoding:
+            encoding = DEFAULT_ENCODING;
+        h = FrameHeader(self.tagHeader);
+        h.id = LYRICS_FID;
+        self.addFrame(LyricsFrame(h, encoding = encoding, lang = lang,
+                                   description = description,
+                                   lyrics = lyrics));
+
    def setUniqueFileIDFrame(self, owner_id, id):
       assert(isinstance(owner_id, str) and isinstance(id, str));
 
@@ -1766,14 +1897,16 @@ class FrameSet(list):
       else:
          raise TypeError("FrameSet key must be type int or string");
 
-#  Mmmmm!  Cheesy!
 def splitUnicode(data, encoding):
-    if encoding == LATIN1_ENCODING or encoding == UTF_8_ENCODING or\
-       encoding == UTF_16BE_ENCODING:
+    if encoding == LATIN1_ENCODING or encoding == UTF_8_ENCODING:
         return data.split("\x00", 1);
-    elif encoding == UTF_16_ENCODING:
-        (d, t) = data.split("\x00\x00\x00", 1);
-        d += "\x00";
+    elif encoding == UTF_16_ENCODING or encoding == UTF_16BE_ENCODING:
+        # Two null bytes split, but since each utf16 char is also two 
+        # bytes we need to ensure we found a proper boundary.
+        (d, t) = data.split("\x00\x00", 1);
+        if (len(d) % 2) != 0:
+            (d, t) = data.split("\x00\x00\x00", 1);
+            d += "\x00";
         return (d, t);
 
 #######################################################################
@@ -1797,6 +1930,9 @@ def createFrame(frameHeader, data):
   # Comment Frames.
   elif COMMENT_FRAME_RX.match(frameHeader.id):
      f = CommentFrame(frameHeader, data = data);
+  # Lyrics Frames.
+  elif LYRICS_FRAME_RX.match(frameHeader.id):
+     f = LyricsFrame(frameHeader, data = data);
   # URL Frames.
   elif URL_FRAME_RX.match(frameHeader.id):
      if USERURL_FRAME_RX.match(frameHeader.id):
@@ -1823,8 +1959,8 @@ def createFrame(frameHeader, data):
 
 
 def map2_2FrameId(originalId):
-    
-    if not TAGS2_2_TO_TAGS_2_3_AND_4.has_key(originalId): return originalId
+    if not TAGS2_2_TO_TAGS_2_3_AND_4.has_key(originalId):
+        return originalId
     return TAGS2_2_TO_TAGS_2_3_AND_4[originalId]
         
         
